/*
   floyd.h for floyd
   Split from floyd.c in preliminary version 1.2beta:
   Tue May 25 23:02:50 1999
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>		/* usleep(), read(), close(), getpid() */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#ifdef linux
#include <linux/kd.h>
#include <sys/ioctl.h>
#else
You lose.
#endif

/* 
   Time values

   LOADAVG1:
   Default LED OFF duration in float at loadavg==1, unit is usec (ulong)
   Default=200000.0: makes 200msec off.
   2 ^ 32 == 4294967296
 */
#define LOADAVG1 200000L
#define MSEC1000 1000000L       /* 1000msec by usec unit */
#define MSEC150  150000L        /* 150msec  */
#define MSEC100  100000L        /* 100msec  */
#define MSEC50   50000L         /* 50msec   */
#define MSEC10   10000L         /* 10msec   */
#define MSEC1    1000L          /* 1msec    */
#define USEC1    1L             /* 1usec    */
#define SEC1     MSEC1000       /* 1 second */

/* Turn into MORSE mode exceeding BIGLOAD_SW */
/*#define MORSE*/               /* Now defined by Makefile/make command line */

#ifdef MORSE
#ifndef BIGLOAD_SW
#define BIGLOAD_SW		/* MORSE requires this. */
#endif
#endif

/* but, this is current default even without MORSE */
#define BIGLOAD_SW

#ifdef BIGLOAD_SW
#define LONGON   (10 * MSEC10)	/* Long `on' duration, unit : usec */
#define SHORTON  (5 * MSEC10)		/* Short `on' LONGON/2, unit : usec */
#define BIGLOAD  8.5            /* Load threshold to switch LONGON/SHORTON */
#endif

#define UPDATECYCLE 4           /* Blink-recalculation cycle (second) */
#define MAXBLINK SEC1		/* Maximum interval, unit: usec*/
#define SLEEP    MSEC50		/* A slight sleep to reduce %CPU ? (usec) */

/* Short sleep used in -p/-E mode
   Shortend temporally (in 1.3.pre05, which may have something wrong
   in response) */
#if 1
#define NAP      MSEC10
#else
#define NAP      (MSEC50/2)
#endif

/* Macros on declarations */
#ifndef __GNUC__
#define inline 
#endif

#if 0				/* Is There different {sys,linux}/types.h ? */
typedef unsigned long ulong;
typedef unsigned char unchar;
#else
#define ulong unsigned long
#define unchar unsigned char
#endif

/*
   Default LED to blink, other possibility is
   LED_NUM: NumLock, ScrollLock: LED_SCR
 */
#ifndef BLINKLED
#define BLINKLED LED_CAP        /* CapsLock, Default of default */
#endif

/*
   Extern globals
 */
extern char progname[];		/** floyd.c **/
extern char version[];

/* VC device */
extern int fdes;		/* Default input file descriptor: 0 */
extern char fname[];		/* Default input filename: stdin */

extern float l1, l2, l3;	/* Load averages, Only l1 is used */
extern time_t currenttime;	/* Should it be extern ? */

/*
   floyd job parameter
 */
typedef struct floyd_param {
  unchar led;			/* LED to blink */

  char debug;
  char verbose;
  char tfloyd;			/* -t : text mode */
  char force;			/* Hidden opt, keep going even in error */

  ulong interval;		/* LED Blinking interval */ 
  ulong maxblink;		/* Maximum Blinking interval */
  ulong fixedsleep;		/* Fixed period added to blink interval */
  ulong ontime;			/* LED `on' duration  */ 
#if 0
  float l1, l2,l3;		/* 1min/5min/15min load averages */
#else
  float *loadavgp;		/* Pointer to l1 (default), l2, or l3  */
#endif
#ifdef BIGLOAD_SW
  float bigloadsw;		/* Threshold to SW LONGONG/SHORTON */
#endif
  float loadavg1;		/* Default blink interval at L.A. 1.0 */
  int updatecycle;		/* /proc/loadavg polling interval */

#ifdef PACKETLOAD		/** packetload.c **/
  char wywhopt;
  char ifname[16];		/* Interface name */
  int iftop;			/* ifs offset in /proc/net/dev */
  int fsize;			/* size of /proc/net/dev */
#endif
#ifdef MORSE
  char morse;			/* Enable MORSE mode in heavy load or not */
#endif
} floyd_param;

typedef struct floyd_job {
#if 0
  jobid id;			/* identifier of this job_param */
#endif
  void *(*fun)();		/* Handler */
  struct floyd_param *param;	/* ptr to its floyd_param */
  struct floyd_job *prev;
} floyd_job;

extern floyd_param gfprm;	/* Grandmaster floyd_param */
extern int do_fork;
extern int ledmodeworking;

/* Extern Functions */
extern void usage();

extern void blink(unchar, ulong, ulong);
extern void ledctrl(unchar, int);
extern void ledalloff();
#ifdef MORSE
extern void Morse(unchar, int);
#endif
/*extern static void lederror();*/

extern char *ledname(floyd_param *fpp);
extern void *shineon(floyd_param *fpp);
extern void *floydloop(floyd_param *fpp);
extern int  voodoo_ctrl(int);
extern void terminator();
extern void showinit(floyd_param *fpp);
extern void main();

extern void register_floyd_job(void * (*func)());
#ifdef PTHREAD
extern void launch_all(int fcmd, int jcmd);
#endif
extern int  find_ledmode();
#if defined(PTHREAD) && defined(PROCTHREAD)
extern void floyd_set_minsleep(ulong);
extern void *floyd_getproc_loop(floyd_param *);
#if defined(PACKETLOAD)
extern void pload_set_minsleep(ulong);
extern void *pload_getproc_loop(floyd_param *);
#endif
#endif
void do_job(int);

#ifdef PACKETLOAD		/** packetload.c **/
#define LISTEN_IFS "eth0"
extern void crazydiamond(floyd_param *fpp);
extern void wywhfiddle();
#endif

/* For extern void launch_all(int fcmd, int jcmd) */
/* fcmd */
#define PARENT_RETURN   0
#define PARENT_EXIT     1

/* jcmd */
#define NO_JOIN         0
#define DO_JOIN         1
/* #define PARENT_CONTINUE 2 */
/* #define PARENT_THREAD   4 */
/* #define PARENT_DEADEND  8 */

#define VOODOO_CHILE    0
#define PARENT_RETURNED 1

/*  
    Compilation controllers
 */
/* #ifndef  FLOYD */
/* #define  FLOYD */
/* #endif   FLOYD */

/* 
 local variables:
 tab-width: 8
 end:
*/
