/*

  jobmgr.c

  Sun Jun 13 12:37:06 1999
  s.n.

  Register all the floyd job, and launch 'em all afterards.

  The compile-command at the bottom expects that `led-inline.c'
  is included in floyd.c and packetload.c.

*/

/*

   Header part

*/
#include "floyd.h"
#include "voodoo.h"

#ifdef TEST_MGR
extern void voodoo_shineon(floyd_param *jp);	 /* In this file */
extern void voodoo_floydloop(floyd_param *jp);	 /* In this file */
#endif

/*
   Code part
 */

/*

   Allocate a struct floyd_param having current struct `gfprm' image

 */

static floyd_job *mjlp = (void *)0; /* Master pointer to job linked list. */
int do_fork   = 0;

floyd_param *alloc_floyd_param() {
  floyd_param *fprmp;

  if ((fprmp = (floyd_param *)malloc(sizeof(floyd_param)))
      == NULL) {
    perror("floyd_param malloc"); exit(errno);
  }
  
  *fprmp = gfprm;

  return fprmp;
}

/*

    Register a struct floyd_job to linked list
    Registering a job has nothing to do with threading.

 */

void register_floyd_job(void * (*func)()) {
  floyd_job *jp;
#ifdef DEBUG
  static int count = 0;
#endif

  if ((jp = (floyd_job *)malloc(sizeof(floyd_job))) == NULL) {
    perror("floyd_job malloc"); exit(errno);
  }
  jp->param = alloc_floyd_param();

  jp->fun = func;

#if 0
  switch (job) {
    case JOB_FLOYD:
      jp->id = JOB_FLOYD;
      break;
    case JOB_CZYDMD:
      jp->id = JOB_CZYDMD;
      break;
  }
#endif

  jp->prev = mjlp;
  mjlp = jp;
#ifdef DEBUG
  fprintf(stderr, "register_floyd_job, count:%d pid:%d\n", ++count, getpid());
  fflush(stderr);
#endif
}

/*

   fork(2) handler

 */

static int xfork(int directive) {
  int pid;

  do_fork = 0; /* We do fork() just one time, but don't rely on this. */

  if ((pid = fork()) > 0) {
    if (gfprm.verbose)
      fprintf(stderr, "%s [%d] forked MORE %s [%d]\n",
	      progname, getpid(), progname, pid);
    if (directive == PARENT_EXIT) exit(0);
    else return PARENT_RETURNED;
  } else if (pid < 0) {
    perror("fork");
    exit(1);
  } /* else go waltzing Matilda with me */
#ifdef DEBUG
  if (gfprm.verbose)
    fprintf(stderr, "%s [%d]: Am child\n", progname, getpid());
#endif
  return VOODOO_CHILE;
}

/*

   The floyd_job linked list inspectors, and add_pload_procthread()

   track_floyd_job(int cmd) -- track down floyd_job linked list.
                               Returns a pointer to floyd_job node on
			       each call. If arg `cmd' is 0, it resets
			       the builtin static pointer to top.
			       Returns (void *)0 at the bottom.

   Thu Jul  8 00:01:34 1999 s.n.

 */

static floyd_job *track_floyd_job(int cmd) {
  static floyd_job *tjlp = (void *)0;
  floyd_job *fjp;
#ifdef DEBUG2
  static int count = 0;

  if (!cmd) count = 0;
  fprintf(stderr, "track_floyd_job: count: %d\n", ++count);
#endif

  if (!cmd) tjlp = mjlp;	/* reset to top of the list */
  
  if (!(fjp = tjlp)) return fjp;
  else tjlp = tjlp->prev;

  return fjp;
}

/* This should be called before launch_all(), which calls reset_floyd_job() */
/* floyd.c needs this: */
/* static */
int find_ledmode() {
  floyd_job *fjp;

  for (fjp = track_floyd_job(0); fjp ; fjp = track_floyd_job(1))
    if (!fjp->param->tfloyd) return 1;

  if (!gfprm.tfloyd) return 1;	/* Now do_job() uses (un-registered) gfprm */
  return 0;
}

#if defined(PROCTHREAD) && defined(PTHREAD)
/* This should be called before launch_all(), which calls reset_floyd_job() */
static int find_floydopt() {
  floyd_job *fjp;

  for (fjp = track_floyd_job(0); fjp ; fjp = track_floyd_job(1))
    if (!fjp->param->wywhopt) return 1;

  if (!gfprm.wywhopt) return 1;	/* Now do_job() uses (un-registered) gfprm */
  return 0;
}

/* 
    add_floyd_procthread() -- Register pload_getproc_loop() only once
 */

static int add_floyd_procthread() {
  static int procthrregistered = 0;

  if (!procthrregistered) {	/* Not registered one yet. */
    if (find_floydopt()) {
      register_floyd_job(floyd_getproc_loop);
      return (procthrregistered = 1);
    }
  }
  return 0;
}

#if defined(PACKETLOAD)
/* This should be called before launch_all(), which calls reset_floyd_job() */
static int find_wywhopt() {
  floyd_job *fjp;

  for (fjp = track_floyd_job(0); fjp ; fjp = track_floyd_job(1))
    if (fjp->param->wywhopt) return 1;

  if (gfprm.wywhopt) return 1;	/* Now do_job() uses (un-registered) gfprm */
  return 0;
}

/* 
    add_pload_procthread() -- Register pload_getproc_loop() only once
 */

static int add_pload_procthread() {
  static int procthrregistered = 0;

  if (!procthrregistered) {	/* Not registered one yet. */
    if (find_wywhopt()) {
      register_floyd_job(pload_getproc_loop);
      return (procthrregistered = 1);
    }
  }
  return 0;
}
#endif /* PACKETLOAD */
#endif /* defined(PROCTHREAD) && defined(PTHREAD) */

#if defined(DEBUG) || defined(TEST_MGR)
void dump_floyd_job() {
  floyd_job *fjp;
  for (fjp = track_floyd_job(0); fjp ; fjp = track_floyd_job(1)) {
    fprintf(stderr, 
	    "FUNCTION: %s\n",
	    (fjp->fun == shineon ? "shineon" :
	     (fjp->fun == floydloop ? "floydloop" :
	      (fjp->fun == pload_getproc_loop ? "pload_getproc_loop":
	       "unknown"))));
    showinit(fjp->param);
  }
}
#endif

#ifdef PTHREAD
/*

   launch_all(fcmd, jcmd) -- Launch all the registered job (without join)
   do_job(fcmd)           -- Call job handler, and loops in it.

   launch_all() uses the master ptr mjlp. 
   do_job() uses gfprm directly.

   If global `do_fork' is non-zero, launch_all() and do_job() forks a child.

   If `fcmd' is one of PARENT_EXIT or PARENT_RETURN which specifies the
   parent's behaviour after fork.

   The child goes on threading in launch_all(), or
   goes into a loop with job handler function in do_job().

   launch_all() calls join_all() to join the threads, if `jcmd` is DO_JOIN.

   Typical use:

     launch_all(PARENT_EXIT, DO_JOIN);  => do all jobs with threading.

     launch_all(PARENT_EXIT, NO_JOIN);  => fire up all thread, but do not join
     registerone();
     do_cmd(PARENT_EXIT);               => goes into loop without `join'

 */

static pthnode *pthlp = (pthnode *)0;

static inline void join_all() {
  /* If Called repeatedly, WITH non-zero dothread
     THIS put master process in WAIT status */
  if (pthlp) voodoo_join(pthlp); 
}

static inline void reset_floyd_job() {
  floyd_job *tjlp, *p;
  for (tjlp = mjlp; tjlp ;) {

    p = tjlp->prev;
    free((void *)tjlp);
    tjlp = p;
  }
  mjlp = (void *)0;		/* Reset mjlp so that register again. */
}

static inline void launch() {
  floyd_job *tjlp;
#ifdef DEBUG
  int count=0;
#endif

  for (tjlp = mjlp; tjlp; tjlp = tjlp->prev) {
#ifdef DEBUG
    fprintf(stderr,
	    "(%d)LAUNCH ifname:%s, %5d [%5d:%5d]\n",
	    do_fork,
	    tjlp->param->ifname,
	    ++count, getpid(), (int)pthread_self());
    fflush(stderr);
#endif
    pthlp = voodoo_create(tjlp->fun, (void *)tjlp->param);
  }

    /* Must free before join.
       Do not free() floyd_param: it is in use.
       */
    reset_floyd_job();
}

void launch_all(int fcmd, int jcmd) {
#ifdef DEBUG
  fprintf(stderr, "launch_all1, pid:%d fcmd:%d jcmd:%d\n",
	  getpid(), fcmd, jcmd);
  fflush(stderr);
#endif

  /* fork(), then child starts threading, parent exits. */
  if (do_fork) {
    if (fcmd == PARENT_EXIT) xfork(PARENT_EXIT);
    else { 
      if (xfork(PARENT_RETURN) == PARENT_RETURNED) return;
    }
  }
  /* Child go waltzing Matilda with me */

  /* Add required procthread */
#if defined(PTHREAD) && defined(PROCTHREAD)
  add_floyd_procthread();
#if defined(PACKETLOAD)
  add_pload_procthread();
#endif
#endif

#ifdef DEBUG
  fprintf(stderr, "launch_all1, pid:%d fcmd:%d jcmd:%d\n",
	  getpid(), fcmd, jcmd);
  fflush(stderr);
#endif

  launch();
  if (jcmd == DO_JOIN) join_all();
}

#endif /* PTHREAD */

void do_job(int fcmd) {

  /* fork(), then child starts threading, parent  returns. */
  if (do_fork) {
    if (fcmd == PARENT_EXIT) xfork(PARENT_EXIT);
    else { 
      if (xfork(PARENT_RETURN) == PARENT_RETURNED) return;
    }
    /* else VOODOO_CHILE must have been returned here */
  }
  
  /* Add required procthread, and launch it if registered */
#if defined(PTHREAD) && defined(PROCTHREAD)
  if (add_floyd_procthread()) launch();
#if defined(PACKETLOAD)
  if (add_floyd_procthread() || add_pload_procthread()) launch();
#endif
#endif

  /* No job is registered for do_job(), which just uses gfprm.
     */
#ifdef PACKETLOAD
  wywhfiddle();
  if (gfprm.wywhopt) shineon(&gfprm);
  else
#endif
    floydloop(&gfprm);

  /*reset_floyd_job();*/		  /* Can't be reached. */
}

/*
   Tester part
 */

#ifdef TEST_MGR
#if 0
int maxcount = 200;
#endif

/* VC device */
int fdes;		/* Default input file descriptor: 0 */
char fname[80];		/* Default input filename: stdin */

char progname[] = "jobmgr";

/* In led.c: BITMAP_LED_CTRL revived #ifdef TEST_MGR */
extern void *os_ifs_loop(void *);

#if 0 /* Unused anymore: now shineon has pload_set_minsleep() call  */
void voodoo_shineon(floyd_param *fpp) {
#if 0
  int count;
  for (count = 0 ; count < 20; count++) {
    printf("(%d)voodoo_shineon ifname:%s %5d [%5d:%5d]\n",
	   do_fork, fpp->ifname, count, getpid(), (int)pthread_self());
    usleep(200000L);
  }
#endif
  pload_set_minsleep(fpp->fixedsleep);
  shineon(fpp);
}

void voodoo_floydloop(floyd_param *fpp){
#if 0
  int count;
  for (count = 0 ; count < maxcount; count++) {
    printf("(%d)voodoo_floydloop ifname:%s %5d [%5d:%5d]\n",
	   do_fork, fpp->ifname, count, getpid(), (int)pthread_self());
    usleep(200000L);
  }
#endif
  pload_set_minsleep(fpp->fixedsleep);
  floydloop(fpp);
}
#endif /* Unused anymore */

void main(int ac, char **av) {

  extern void showinit();
  extern void registerone();
  extern int sleepopt;

  floyd_job *tjlp;

  if (ac > 1) {
    sscanf(av[1], "%d", &do_fork);
  }

#if 0
  if (ac > 2) {
    sscanf(av[2], "%d", &maxcount);
  }
#endif

  signal(SIGINT, terminator);
  signal(SIGHUP, terminator);
  signal(SIGTERM, terminator);

  printf("1ledmodeworking:%d\n", ledmodeworking);
  ledmodeworking = find_ledmode();
  printf("2ledmodeworking:%d\n", ledmodeworking);

  /* Trying to set scheduling parameters for os_ifs_loop(), but
       1) This method requires suid root.
       2) Something is still wrong in sched/sched.c...
     Perhaps I won't use scheduling param modification in pfloyd.
     Sounds like it is too hairy. Mon Jun 28 13:33:06 1999 s.n.
   */
#if 0 /* something is wrong: freeze the whole system... */
  {
    /* these are in sched/sched.c */
    extern pthread_attr_t *voodoo_modsched (pthread_attr_t *thr_attr);
    pthread_attr_t thr_attr;
    pthread_t thr;

    voodoo_modsched(&thr_attr);	/* in sched/sched.c */

    /* voodoo_create((void *)os_ifs_loop, 0);*/
    pthread_create(&thr, &thr_attr, (void *)os_ifs_loop, 0);
  }
#endif

  gfprm.tfloyd =  0;
  gfprm.verbose =  0;

  gfprm.led =  LED_CAP;
  gfprm.debug =  0;
  gfprm.wywhopt =  0;
  gfprm.fixedsleep =  50000L;	/* 50msec in usec */
  registerone();

  gfprm.led =  LED_NUM;
  gfprm.debug =  0;
  gfprm.wywhopt =  1;
  strcpy(gfprm.ifname, "ul0");
  sleepopt = 1;
  gfprm.fixedsleep =  9000L;	/* 9msec in usec */
  registerone();

  gfprm.led =  LED_SCR;
  gfprm.debug =  0;
  gfprm.wywhopt =  1;
  strcpy(gfprm.ifname, "eth0");
  gfprm.fixedsleep =  10000L;	/* 10msec in usec */
  registerone();

#if 1
  printf("(%d)Parent: %d\n", getpid(), do_fork);

  for (tjlp = mjlp; tjlp ;tjlp = tjlp->prev) {
    printf("jobmgr-parent ifname:%s debug:%d verbose:%d\n",
	   tjlp->param->ifname, tjlp->param->debug, tjlp->param->verbose);
  }
#endif

  /* Now launch_all() or do_job() calls add_pload_procthread() AND/OR
     add_floyd_procthread(), which automatically fires up one
     required `procthread' loop as a thread.  Not required here anymore: */
  /*   register_floyd_job(pload_getproc_loop); */

#if 1 /* track_floyd_job() test */
  dump_floyd_job();
#endif

  launch_all(PARENT_EXIT, DO_JOIN);
}
#endif /* TEST_MGR */
/*
;; Jus test:
;;       packetload.c \\
;;      sched/sched.o \\
;;   -DDEBUG2    -DDEBUG \\
;;   -DDEBUG \\
;; Make TEST version (no -DFLOYD)
  local variables:
  compile-command: "make clean ; cc -s -static -Wall \\
 -D_REENTRANT -DTEST_MGR -DPACKETLOAD -DPTHREAD -DPROCTHREAD \\
-I /hdawork/src/glibc2/glibc-2.0.7pre6/linuxthreads/sysdeps/pthread \\
-o jobmgr jobmgr.c voodoo.c led.c    floyd.c \\
              packetload.procthread.c \\
-L /hdawork/src/glibc2/glibc-2.0.7pre6/linuxthreads -lpthread \\
-L /hdawork/src/glibc2/glibc-2.0.7pre6 -lc"
  end:
*/
