/*
   led.c for floyd

   Sat Jun  5 21:08:48 1999
   Splitted from floyd.c. s.n.
 */

#include "floyd.h"

/*

   BITMAP control version of led controllers

 */
#ifdef TEST_MGR
#undef BITMAP_LED_CTRL		/* Currently unused */
#endif
#ifdef BITMAP_LED_CTRL

/* This version just call
   ioctl() using `led' value.
   extern oleds, nleds are in led.c, floyd.h has these.
 */

char oleds, nleds;

void ledctrl0() {

  if (ioctl(fdes, KDGETLED, &oleds)) {
    perror("KDGETLED"); goto quit;
  }

  if (oleds != nleds) {
    /*fprintf(stderr, "ledctrl0: %d:%d\n", oleds, nleds);*/
    if (ioctl(fdes, KDSETLED, nleds)) {
      perror("KDSETLED"); goto quit;
    }
  }
  return;
 quit:
  fprintf(stderr,
          "LED status write error... Careful With That %s, Eugene,\n",
          fname);
  close(fdes);
  exit(1);
}

/* New version of ledctrl().
   Just bitmap operation according to led l.
   extern oleds, nleds are in led.c, floyd.h has these.
 */

void ledctrl(unchar l, int onoff) {
  unchar nval, ndef;

  /*fprintf(stderr, "ledctrl: l:%d onoff:%d\n", l, onoff);*/
  nval = ndef = 0;
  ndef |= l;
  nval |= onoff ? l : 0;
  nleds = (oleds & ~ndef) | nval;
}

/* A simple loop calling ledctrl0() repeatedly
 */

ulong ioctlwait = 400L;		/* 400us: too fast ? */
/*ulong ioctlwait = 800L;*/	/* 800us */
/*ulong ioctlwait = 3000L;*/	/* 3ms too slow */
/*ulong ioctlwait = (MSEC50/2);*//* 25ms: wywh/floyd won't blink nicely */

void *os_ifs_loop(void *arg) {
  fprintf(stderr, "os_ifs_loop firedup [%d]\n", getpid());
  for (;;) {
    /*fprintf(stderr, "%s \n", (char *)arg);*/
    ledctrl0();
    usleep(ioctlwait);
  }
}
#else /* BITMAP_LED_CTRL */

char force = 1;			/* Do not exit on LED ioctl() (for a while) */
static void lederror() {
  if (gfprm.force) {		/* Potential bug */
    char ct[64];
    time_t t; 

    if (gfprm.verbose) {	/* Potential bug */
      time(&t);
      strcpy(ct, ctime(&t));
      *(ct + (strlen(ct) - 1)) = '\0';
      fprintf(stderr,
	      "%s %s [%d]: led access error %s(fd:%d)\n",
	      ct, progname, getpid(),
	      *fname ? fname : "stdin", fdes);
    }
    return;
  } else {
    fprintf(stderr,
	    "%s [%d]: led access error... careful with that %s(fd:%d), Eugene,\n"
	    , progname, getpid(), *fname ? fname : "stdin", fdes);
    close(fdes);		/* Use default 0 at the top of main() */
    exit(1);
  }
}

void ledctrl(unchar l, int onoff) {
  unchar nleds;			/* LED variable for ioctl() */
  unchar oleds;			/* LED variable for ioctl() */
  char nval, ndef;

#ifdef linux
  if (ledmodeworking) {
    if (ioctl(fdes, KDGETLED, &oleds)) {
      perror("KDGETLED"); lederror();
    }

    nval = ndef = 0;
    ndef |= l;
    nval |= onoff ? l : 0; 
    nleds = (oleds & ~ndef) | nval;

    if (oleds != nleds) {
      if (ioctl(fdes, KDSETLED, nleds)) {
	perror("KDSETLED"); lederror();
      }
    }
  } else {
    fprintf(stderr, (onoff ? "***\r" : "   \r"));
  }
#else
  Currently only for linux. Add your own code here for your OS.
#endif
}
#endif /* BITMAP_LED_CTRL */

void blink(unchar l, ulong on, ulong off) {
  ledctrl(l, 1);
  usleep(on);
  ledctrl(l, 0);
  usleep(off);
}

/* aka. ledcheck() */
void ledalloff() {
  if (ioctl(fdes, KDSETLED, 0)) {
    perror("KDSETLED");
    fprintf(stderr,
"%s [%d]: ledalloff: led access error... careful with that %s(fd:%d), Eugene,\n"
	    , progname, getpid(), *fname ? fname : "stdin", fdes);
    exit(1);
  }
}

#ifdef MORSE
void Morse(unchar l, int ontimes) {
  register ulong shorton  = (MSEC150); /* 150ms */
  register ulong shortoff = (MSEC150); /* 150ms */
  register ulong longoff  = (8 * MSEC100); /* 800ms */
  
#ifdef MORSE_DEBUG
  fprintf(stderr, "Morse: ontimes:%d.\n", ontimes);
#endif

  /* Short blinks X ontimes */
  for (; ontimes > 0 ; ontimes--) {
    blink(l, shorton, shortoff);
  }
  /* Long off */
  usleep(longoff - shortoff);
}
#endif
/*
  local variables:
  compile-command: "make DEBUGDEF='-DDEBUG -DBTC5100' floyd"
  tab-width: 8
  c-indent-level: 2
  c-continued-statement-offset:  2
  c-brace-offset: 0
  c-argdecl-indent: 0
  c-label-offset: 2
  end:
*/
