/*

   packetload.c for floyd

   Basic code written by CXA01035@nifty.ne.jp (Zophos)
   (c) 1999 CXA01035@nifty.ne.jp (Zophos)
  
   NOTES:
   Tue Jul  6 12:55:00 1999
   /proc/net/dev accessing routine is launched as a thread.

   Mon Jun 21 14:05:23 1999
   Now iftop/fsize are in floyd_param. They should be independent each
   other in all the threads.

   Mon Jun  7 12:59:27 1999
   Tuned getstatics() and crazydiamond() somemore. Search today's date
   hereunder for the relevant comments and codes.
   s.n.

   Wed Jun  2 21:50:52 1999
   Lot of changes. Almost rewritten all over based on Zophos patch on
   99/05/30.  Function/variable names are changed.  Duplicated
   /proc/net/dev accessing part is removed.  Codes passing around
   pointers thought to be of extreme hairiness and redundancy are
   removed if it is possible. The gettop() has verbose option to 
   report comming and going of the relevant interface.
   s.n.

   Thu May 27 19:23:48 1999
   In the evening of 99/05/24, Zophos sent the code in a patch form
   for floyd-1.2alpha, and I splitted it into this file for floyd-1.2beta.
   s.n.

   TODO: (almost done ?)
   Support ppp0, ul0, dummy0 or any other interface.
   Do not raise error even if ppp0/ul0/dummy0 interfaces are dynamically
   attached or detached.
*/
#include "floyd.h"

#ifdef PACKETLOAD

/*
	get kernel version -- 2nd digit of kernel version is returned as int.
        Found that the /proc/net/dev format is totally different in 2.0.x
        and 2.2.x.
        Wed Jun 16 03:02:30 1999 s.n.
 */

int kversion(/*char *s*/) {
#ifdef linux
  static int kernelv = -666;
  int fd, len;
  char buffer[256];
  char vstring[14];
  char *p;

  if (kernelv >= 0) return kernelv;

  /*if (s) strcpy(vstring, s);
  else*/
  {
    fd = open("/proc/version", O_RDONLY);
    len = read(fd, buffer, sizeof(buffer) - 1);
    close(fd);
    buffer[len] = '\0';

    for(p = buffer; *p && !('0' <= *p && *p <='9'); p++) {} /* get figure */
    sscanf(p, "%s", vstring);
  }

  /*
    Watch that bug, Eugene !
    Order is significant.  Do not support 1.x.x kernel
    Must add lot more lines when 2nd field reached 2digits.

    $ ./kversion 2.12.999999999999 ; echo $?     # See 00test/kversion.c
    1
    $ ./kversion 2.23.999999999999 ; echo $?
    2
  */
  if (strcmp(vstring, "2.7.0") >= 0) return (kernelv = 7);
  if (strcmp(vstring, "2.6.0") >= 0) return (kernelv = 6);
  if (strcmp(vstring, "2.5.0") >= 0) return (kernelv = 5);
  if (strcmp(vstring, "2.4.0") >= 0) return (kernelv = 4);
  if (strcmp(vstring, "2.3.0") >= 0) return (kernelv = 3);
  if (strcmp(vstring, "2.2.0") >= 0) return (kernelv = 2);
  if (strcmp(vstring, "2.1.0") >= 0) return (kernelv = 1);
  if (strcmp(vstring, "2.0.0") >= 0) return (kernelv = 0);
  /* else */
  return (kernelv = -1);
#else
  You lose
#endif
}

/*
	Was readstart() -- Get Starting point of /proc/net/dev
	 (by Zophos)

        Found|Can't be found|Disappeared message is not supported in PTHREAD.
        The cost of the messaging is too high.
        s.n. Tue Jun 22 16:20:09 1999

        Modified just to scan char buffer.
        s.n. Wed Jun  2 21:50:52 1999
*/

#if defined(PTHREAD) && defined(PROCTHREAD)
#include "voodoo.h"
static pthread_mutex_t pload_proc_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
#undef STRICT

static inline int gettop(int len, char *buffer,
                         char *interface, floyd_param *fpp) {
	register int i,j;
        register int result;    /* result/oresult should be int. */
#ifdef PTHREAD
#else
        static int oresult=-2;
#endif
#if !defined(STRICT) && defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
#endif
	for(i=0,j=0;(i<len)&&(interface[j]!='\0');i++){
		if(buffer[i]==interface[j]){
			j++;
		}
		else{
			j=0;
		}
	}
        if (j > 0) result=i+1;  /* Colon skipped, result: the least is 1 */
        else result=-1;         /* No such interface. */
#if !defined(STRICT) && defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_unlock (&pload_proc_mutex) != 0) {
                perror("Unlock mutex"); exit(1);
        }
#endif

#ifdef PTHREAD
#else
        if (fpp->verbose) {          /* s.n. or (fpp->debug) ? */
                if (oresult < 0 && result > 0)
                        fprintf(stderr, "%s [%d]: interface %s found.\n",
                                progname, getpid(), interface);
                else if (result < 0)
                        fprintf(stderr, "%s [%d]: interface %s %s.\n",
                                progname, getpid(), interface,
                                oresult == -2 ?
                                "can't be found" : "disappeared");
                oresult = result;       /* Remember this */
        }
#endif
        return result;
}

/*

   If defined(PROCTHREAD), /proc/net/dev is accessed by
   an independent thread.  The `read' buffer (rawbuffer) is shared by
   packetload mode thread(s).

   Mon Jul  5 13:10:05 1999

 */

/*
   static inline getproc(char *fn)
   Used also in non-PTHREAD, or non-PROCTHREAD version.

   Perhaps mutex_lock is required (seems like working without it though).
 */

static char rawbuffer[1500];
static int len = -666;          /* A sinister value */
static inline void getproc(char *fn) {
	int fd;

#ifdef DEBUG
        fprintf(stderr, "%s [%d]: getproc(), len:%d.\n",
                progname, getpid(), len);
#endif
	fd = open(fn, O_RDONLY);
#if defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
	len = read(fd, rawbuffer, sizeof(rawbuffer) - 1);
	close(fd);
	rawbuffer[len] = '\0';
        if (pthread_mutex_unlock (&pload_proc_mutex) != 0) {
                perror("Unlock mutex"); exit(1);
        }
#else
	len = read(fd, rawbuffer, sizeof(rawbuffer) - 1);
	close(fd);
	rawbuffer[len] = '\0';
#endif
}

#if defined(PTHREAD) && defined(PROCTHREAD)
/*
   The /proc/net/dev accessing thread must be looped
   with a delay (minsleep/2).
   The minsleep should be determined by calling 
   pload_set_minsleep(fpp->fixedsleep)
   from the init part of floydloop() or shineon().

   Perhaps mutex_lock is required (seems like working without it though).
*/
#define P_MINSLEEP_DIVISION 2
static long minsleep = -1;      /* Needs to be negative. */

void pload_set_minsleep(ulong x) {
        ulong tem;

        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
        if (minsleep < 0)
                minsleep = x / P_MINSLEEP_DIVISION;
        else if (minsleep > (tem = x / P_MINSLEEP_DIVISION)) {
                minsleep = tem;
        }
        if (pthread_mutex_unlock (&pload_proc_mutex) != 0) {
                perror("Unlock mutex"); exit(1);
        }
}

/*
   A thread routine to read /proc/net/dev
   Currently this is registered first prior to all the packetload threads,
   and threads are fired up in FILO order, so I think this thread should
   have the last pid among the packetload threads, but it's not always. Why.
 */

void *pload_getproc_loop(floyd_param *fpp) {

        if (fpp->debug || fpp->verbose)
                fprintf(stderr, "pload_getproc_loop [%d]\n--\n", getpid());

        for (;;) {
#ifdef linux
                getproc("/proc/net/dev");
#else
you lose
#endif
#ifdef DEBUG
                /*fprintf(stderr, "PLOAD_GETPROC_LOOP: buf:%s\n", rawbuffer);*/
                fprintf(stderr, "PLOAD_GETPROC_LOOP: minsleep: %s\n", 
                        (minsleep < 0 ? "YET" : "DETERMINED"));
                fprintf(stderr, "PLOAD_GETPROC_LOOP: minsleep: %ld\n", 
                        (minsleep < 0 ? 5000L : minsleep));
#endif
                usleep(minsleep < 0 ? 5000L : minsleep);
        }
        return (void *)0;       /* Just to be PTHREAD compliant */
}
#endif /* defined(PTHREAD) && defined(PROCTHREAD) */

/*
	Was getpacketnum() -- Get Packet Number
	 (by Zophos)

        Mon Jun 21 14:05:23 1999
        Now iftop/fsize are in floyd_param:
        (progn
          (save-excursion
            (query-replace-regexp "\\biftop\\b" "fpp->iftop"))
          (save-excursion
            (query-replace-regexp "\\bfsize\\b" "fpp->fsize")))

        Mon Jun  7 13:50:14 1999
        Passing kernelversion as 2nd arg: the implicit expectation is that
        getstatics() would be expanded `inline'.

kernel 2.2.10's /proc/net/dev looks like (annoyingly long):
Inter-|   Receive                                                |  Transmit
 face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed
    lo:  285157     586    0    0    0     0          0         0   285157     586    0    0    0     0       0          0
  eth0:119557305  321118    0    0    0     0          0       348 130237153  298301    0    0    0    24       0          0
  ppp0: 2089019    1444  115    0    0   115          0         0    68031    1029    0    0    0     0       0          0
   ul0:       0     167    0    0    0     0          0         0        0     145    0    0    0     0       0          0

kernel 2.0.36's /proc/net/dev:
Inter-|   Receive                  |  Transmit
 face |packets errs drop fifo frame|packets errs drop fifo colls carrier
    lo:     20    0    0    0    0       20    0    0    0     0    0
  eth0:     32    0    0    0    0       28    0    0    0     0    0
   ul0:     77    0    0    0    0       68    0    0    0     0    0

        Copying to tmp[] (as well as its index int j) removed. The buffer[] is
        used instead.  Perhaps this may be a slight improvement in thruput.
        Now `iftop' and `fsize' are static int within this (and never
        referred from anywhere else).  Having them as extern/static
        adds 30bytes or so to the binary size (5bytes or so bigger in
        x86 if static within getstatics(), not placed outside of the
        function), and some overhead to access these in .DATA segment,
        but pushing two pointers on stack at every call of getstatics()
        is eliminated. Thus, as for thruput, there may not be any
        improvement concerning the `iftop' and `fsize' relocation ?

        In the test after I wrote above, the modification was found to take
        effect. %CPU extremely reduced.

        Is pushing arguments in function-call is that heavy ?
        See ChangeLog for detail....

        # ./floyd: NEW, /usr/local/bin/floyd: OLD.
        USER       PID %CPU %MEM  SIZE   RSS TTY STAT START   TIME COMMAND
        sn        1162  3.4  0.4   828   292   1 S    13:42   0:08 ./floyd -s10 -ECD
        sn        1177 15.1  0.4   828   276   1 S    13:42   0:37 /usr/local/bin/floyd -s10 -END

 */

static inline long getstatics(char *interface, int kernelversion,
                              floyd_param *fpp) {
	long rpn,tpn,dummy;
	int  i;
#ifdef STRICT
        char buffer[1500];
#else
        char *buffer;
#endif

#if defined(PTHREAD) && defined(PROCTHREAD)
        /* In PTHREAD, getproc() is looped by an independent thread.
           so, NOP here.
         */
        if (len == -666) {      /* the procthread not working yet. */
                /* fprintf(stderr, "The procthread not working yet\n"); */
                return -1;
        } /* else fprintf(stderr, "The procthread WORKING !\n"); */
#else
#ifdef linux
        getproc("/proc/net/dev");
#else
	Currently only for linux. Add your own code here for your OS;
#endif
#endif

        /*    Current gettop() and getstatics() seems to work nice
              without copying rawbuffer[] to buffer[]. */
#if defined(STRICT)
#if defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
#endif
        strcpy(buffer, rawbuffer);
#if && defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_unlock (&pload_proc_mutex) != 0) {
                perror("Unlock mutex"); exit(1);
        }
#endif
#else
        buffer = rawbuffer;
#endif

        /* 
           If length of /proc/net/dev has changed,
           must redefine start point. (990530.Zophos)
           Modified to hand buffer, Wed Jun  2 22:10:35 1999 s.n.
        */
        /* If lock the timing is moved up here, 2 shineon threads deadlocks.
           Why ? */
#undef BAD_TIMING  /* This makes a binary lets two shineon threads hang. */
#ifdef BAD_TIMING
#if !defined(STRICT) && defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
#endif
#endif
	if(fpp->fsize != len){                                       /* ^ */
                fpp->fsize = len;                                    /* | */
		fpp->iftop = gettop(len, buffer, interface, fpp);    /* | */
		if(fpp->iftop < 0) return -1;                        /* | */
	}                                                            /* | */
                                                                     /* | */
#ifndef BAD_TIMING
#if !defined(STRICT) && defined(PTHREAD) && defined(PROCTHREAD)      /* | */
        if (pthread_mutex_lock (&pload_proc_mutex) != 0) {
                perror("Lock mutex"); exit(1);
        }
#endif
#endif
	for(i=fpp->iftop; (i<len)&&(buffer[i]!='\n') ;i++) {}
	buffer[i]='\0';
#if 0
        fprintf(stderr, "2: iftop:%d fsize:%d len:%d \n",
                fpp->iftop, fpp->fsize, len);
#endif
        rpn=tpn=0;         /* Perhaps not required, gcc inits them with 0 ? */
        if (kernelversion >= 2)       /* don't support 1.x.x(/2.1.x) */
                sscanf(buffer+fpp->iftop,
                       "%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld",
                       &dummy,&rpn,&dummy,&dummy,&dummy,&dummy,&dummy,&dummy,
                       &dummy,&tpn,&dummy,&dummy,&dummy,&dummy,&dummy,&dummy);
        else
                sscanf(buffer+fpp->iftop,
                       "%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld%ld",
                       &rpn,&dummy,&dummy,&dummy,&dummy,
                       &tpn,&dummy,&dummy,&dummy,&dummy,&dummy);
#if !defined(STRICT) && defined(PTHREAD) && defined(PROCTHREAD)
        if (pthread_mutex_unlock (&pload_proc_mutex) != 0) {
                perror("Unlock mutex"); exit(1);
        }
#endif
	return(rpn+tpn);
}

/*
	crazydiamond(): the only entry point -- external function

	In Zophos patch, this function was a braced block in main().

        Mon Jun  7 13:02:28 1999
        The crazydiamond() is not curious about `iftop' and `fsize'
        anymore, which are hidden in getstatics() as a static int, and
        only getstatics() cares them.

        Thu Jun  3 14:19:29 1999
        ledstate added to reduce times of ledctrl() call.
        Seems like for(;;) loop speeded, but %CPU of floyd goes up due to
        shortening of the getstatics() call cycle.
        s.n.

        Wed Jun 16 03:19:09 1999
        kernel version detection.  See kversion() above.
        I don't know what happens in 1.x.x and 2.1.x kernels...
        s.n.
        
 */

void crazydiamond(floyd_param *fpp) {
	register int ledstate=0;/* force off state */

        register unchar l = fpp->led; /* Copy to register, or stack. */

        register unsigned int count = 0; /* for fpp->debug */
	register ulong xsleep = fpp->fixedsleep;
        register int kv = -1;   /* Kernel version */
	long cpn,opn;

        kv = kversion(/*0*/);   /* /proc/net/dev format is totally different */

        /* opn = getstatics(fpp->ifname, kv);*/
        opn = -1;               /*  Now getstatics is inline maybe. */

	for (;;) {
		cpn = getstatics(fpp->ifname, kv, fpp);
		if(cpn != opn){ /* Must turn it ON */
                        if (fpp->debug)
                                fprintf(stderr,
"shineon:  ifname:%s state:%d count:%8d opn:%ld cpn:%ld diff:%ld\n",
                                        fpp->ifname,
                                        ledstate,
                                        count++, opn, cpn, cpn-opn);
                        if (!ledstate) /* OFF now ? */
                        {       /* state does not matter */
                                ledctrl(l, 1);
                                ledstate++;
                        }
                        else    /* Already ON, NOP or flash */
                        {       
#undef FLASHING
#ifdef FLASHING                 /* Give one blink when ON-ON: non-default */
#define CUTOFF (5*1000L)
                                ledctrl(l, 0);
				usleep(CUTOFF);
                                ledctrl(l, 1);
#endif
                        }
			opn=cpn;
		} else {        /* Or, must turn it off */
                        if (fpp->debug && fpp->verbose) /* This &&: TEMPORARY FEATURE! */
                                fprintf(stderr,
"shineoff: ifname:%s state:%d count:%8d opn:%ld cpn:%ld\n",
                                        fpp->ifname,
                                        ledstate, count++, opn, cpn);
                        if (ledstate)
                        {
                                ledctrl(l, 0);
                                ledstate &= 0;
                        }
		}
		/* usleep(MSEC10); */
		/* Default sleep is trimmed to NAP (20msec default), 
                   unless specifed with -S option.
                   Probably this xsleep test is what the user want. */
		if (xsleep > 0) usleep(xsleep);
	}
}
#endif /* PACKETLOAD */

/*
  "local-variables" section is derived from
  /usr/src/linux-2.0.36/fs/proc/scsi.c (file 2 of 2) line 163/223 (END):
  and changed to match with Zophos style, but it is not perfect yet.
  What is set to 5 ?  Let's checkout c-style-alist later...

  local variables:
  compile-command: "make DEBUGDEF=-DDEBUG floyd"
  tab-width: 8
  c-indent-level: 8
  c-brace-imaginary-offset: 0
  c-brace-offset: -4
  c-argdecl-indent: 4
  c-label-offset: -4
  c-continued-statement-offset: 8
  c-continued-brace-offset: 0
  indent-tabs-mode: nil
  c-basic-offset: 8 
  end:
  ;; Some of offsets seems like not working in 19.34/20.x. Use this:
  eval: (progn (c-set-offset 'label '+ nil) (c-set-offset 'case-label '+ nil))
  ;;tab-width: 2 (this is not what I want)

  Some of above are drived /usr/src/linux/Documentation/CodingStyle.

  Or try eval manually:
  (c-set-style "linux") ;; idealistic K&R is available only in emacs-20.x
  (c-set-style "K&R")   ;; loose K&R, available in emacs-19.34
  Function c-set-style can't be used in emacs-19.28.
 */
